// shopbot.cpp ackmed@gotwalls.com
#include "shopbot.h"
#include "item.h"
#include "packethelper.h"
#include "config.h"
#include "external/LinkedList/LinkedList.h"

char ConfigPath[_MAX_PATH];  // dir where to find ini file
FUNCTIONENTRYPOINTS	*server;
WORD Debug;
char DebugBuffer[1024];
WORD state;
LinkedList *path = new LinkedList;
LinkedList *config = new LinkedList;

NPCSTRUCT	npc;
WPSTRUCT	wp;

BOOL Recording = false;
BOOL SetNPC = false;
BOOL HaveSync = false;
DWORD tick = 0;
LinkedItem_t *currentStep;

// module info
DWORD	ModuleVersion=MAKELONG(VERSION_MAJOR,VERSION_MINOR);
char	ModuleAuthor[]="ackmed";
char	ModuleWebsite[]="www.gotwalls.com";
char	ModuleDescription[]="shopbot for Diablo2 LOD";
char	ModuleEmail[]="ackmed@gotwalls.com";

// module commands
MODULECOMMANDSTRUCT ModuleCommands[]= {
	{	
		"debug",
		OnCommandDebug,
		"debug [devel|packet]c0 toggle debug mode.\n"
		"[devel] adds extra debugging\n"
		"[packet] adds packet level debugging"
	},
	{
		"record",
		OnCommandRecord,
		"recordc0 toggles starting and stoping of path recording"
	},
	{
		"clear",
		OnCommandClear,
		"clearc0 removes the recorded path"
	},
	{
		"reload",
		OnCommandReload,
		"reloadc0 reloads ini file"
	},
	{
		"setnpc",
		OnCommandSetNPC,
		"setnpcc0 used to set which npc to tade with.\n"
		"run this command then click on the npc."
	},
	{
		"help",
		OnCommandHelp,
		"setnpcc0 help!?"
	},
	{
		"start",
		OnCommandStart,
		"startc0 start the bot"
	},
	{
		"stop",
		OnCommandStop,
		"stopc0 stop the bot"
	},
	{NULL}
};

// module command functions
BOOL PRIVATE OnCommandDebug(char **argv, int argc) {

	// no params passed, just do a simple toggle
	if(argc == 2){
		if(Debug) {
			server->GamePrintString("Debugging disabled");
			Debug = DEBUG_OFF;
		} else {
			server->GamePrintString("Debugging enabled");
			Debug = DEBUG_USER;
		}
		return true;
	} else if(argc == 3) {
		if(strcmpi(argv[2],"devel") == 0) {
			server->GamePrintString("Devel Debugging enabled");
			Debug = DEBUG_DEVEL;
			return true;
		} else if(strcmpi(argv[2],"packet") == 0) {
			server->GamePrintString("Packet Debugging enabled");
			Debug = DEBUG_PACKET;
			return true;
		}
	}
	return false;
}


BOOL PRIVATE OnCommandRecord(char **argv, int argc) {
	if(Recording) {
		server->GamePrintString("Path Recording Stopped");
	} else {
		server->GamePrintString("Path Recording Started");
	}
	// toggle it
	Recording = 1 - Recording;
	return true;
}

BOOL PRIVATE OnCommandClear(char **argv, int argc) {
	LinkedItem_t *step;
	if(!Recording) {
		// delete path
		for(step = path->GetFirstItem();step != NULL;step = path->RemoveItem(step)){
			delete (LOCATIONSTRUCT*) step->lpData;
		}
		server->GamePrintString("Deleting Recorded Path");
	} else {
		server->GamePrintString("Cannot Delete Path when Recording");
	}
	return true;
}

BOOL PRIVATE OnCommandReload(char **argv, int argc) {
	LinkedItem_t *cfg;

	// delete the old config
	for(cfg = config->GetFirstItem();cfg != NULL;cfg = config->RemoveItem(cfg)){
		delete (CONFIGSTRUCT*) cfg->lpData;
	}

	// load new config
	if(!LoadConfig()) {
		server->GamePrintString("Error loading Config");
	} else {
		server->GamePrintString("shopbot.ini reloaded");
	}
	return true;
}

BOOL PRIVATE OnCommandSetNPC(char **argv, int argc) {
	SetNPC = true;
	server->GamePrintString("SetNPC Enabled.  Go click on NPC to shop with");
	return true;
}

BOOL PRIVATE OnCommandStart(char **argv, int argc) {
	state = STATE_WPFROMTOWN;
	server->GamePrintString("Starting shopbot");
	return true;
}

BOOL PRIVATE OnCommandStop(char **argv, int argc) {
	state = STATE_NONE;
	server->GamePrintString("Stopping shopbot");
	return true;
}
BOOL EXPORT OnClientJoinGame() {
	LinkedItem_t *step;
	// delete paths that may already exist
	for(step = path->GetFirstItem();step != NULL;step = path->RemoveItem(step)){
		delete (LOCATIONSTRUCT*) step->lpData;
	}
	return true;
}
// functions called by d2hackit
BOOL EXPORT OnClientStart() {
	return LoadConfig();
}
// outgoing packet
DWORD EXPORT OnGamePacketBeforeSent(BYTE* aPacket, DWORD aLen) {
	// dump packet if packet debugging is on
	if(Debug == DEBUG_PACKET) {
		sprintf(DebugBuffer,"%s","SENT: ");
		for(int i = 0;i < aLen;i++) {
			sprintf(DebugBuffer,"%s %.2x",DebugBuffer,aPacket[i]);
		}
		server->GamePrintInfo(DebugBuffer);
	}

	// if recording save any movement 
	if(Recording) {
		if(aPacket[0] == 0x03) {
			LOCATIONSTRUCT *step = new LOCATIONSTRUCT;
			memcpy(&step->x,&aPacket[1],2);
			memcpy(&step->y,&aPacket[3],2);
			if(Debug == DEBUG_DEVEL) {
				sprintf(DebugBuffer,"Recording Step: %d, %d",step->x,step->y);
				server->GamePrintString(DebugBuffer);
			}
			path->AddItem(step);
		}
	}

	// setting npc
	if(SetNPC) {
		if(aPacket[0] == 0x13 && aPacket[1] == 0x01) {
			memcpy(&npc.npcID,&aPacket[5],4);
			npc.pos.x = 0;
			npc.pos.y = 0;
			SetNPC = false;
			if(Debug) {
				sprintf(DebugBuffer,"NPC set to id: %d",npc.npcID);
				server->GamePrintString(DebugBuffer);
			} else {
				server->GamePrintString("NPC is now set");
			}
		}
	}
	// deny client sync with server while bot is running
	if(aPacket[0] == 0x5f && state != STATE_NONE) {
		if(Debug) {
			server->GamePrintString("Denying position sync");
		}
		return 0;
	}
	return aLen;

}

// incoming packet
DWORD EXPORT OnGamePacketBeforeReceived(BYTE* aPacket, DWORD aLen) {

	// dump packet if packet debugging is on
	if(Debug == DEBUG_PACKET) {
		sprintf(DebugBuffer,"%s","RECV: ");
		for(int i = 0;i < aLen;i++) {
			sprintf(DebugBuffer,"%s %.2x",DebugBuffer,aPacket[i]);
		}
		server->GamePrintInfo(DebugBuffer);
	}

	if(aPacket[0] == 0x96) { // location sync from server
		if(Debug) {
			BitFields iPacket(aPacket,aLen);
			iPacket.GetField(23);
			LOCATIONSTRUCT loc;
			loc.x = iPacket.GetField(16);
			loc.y = iPacket.GetField(16);
			sprintf(DebugBuffer,"Location Sync from server (%d,%d)",loc.x,loc.y);
			server->GamePrintInfo(DebugBuffer);
		}
		HaveSync = true;
	} else if(aPacket[0] == ITEM_MESSAGEID_DROP) {
		ITEMSTRUCT *item;
		item = ParseItemDrop(aPacket,aLen);
		if(Debug && item != NULL) {
			DumpItemStruct(item);
		}
		// check for match needs to go here
		if(item != NULL) {
			CONFIGSTRUCT *cfg;
			cfg = ItemMatchCheck(item);
			if(cfg != NULL) {
				// we have a match, stop the bot.
				state = STATE_NONE;
				// print stuff if description is defined.
				if(cfg->Description[0] != '\0') {
					server->GamePrintString(cfg->Description);
				}
			} 
			delete item;
		}
	} else if(aPacket[0] == 0x6d) {  // npc update
		if(npc.npcID != 0) {
			WORD npcID;
			memcpy(&npcID,&aPacket[1],4);
			// see if this is our npc
			if(npc.npcID == npcID) {
				memcpy(&npc.pos.x,&aPacket[5],2);
				memcpy(&npc.pos.y,&aPacket[7],2);
				if(Debug == DEBUG_DEVEL) {
					sprintf(DebugBuffer,"Got NPC loc update (%d,%d)",npc.pos.x,npc.pos.y);
					server->GamePrintString(DebugBuffer);
				}
			}
		}
	} else if(aPacket[0] == 0x51) {	// waypoint info
		if(aPacket[12] == 0x02) {
			memcpy(&wp.entityID,&aPacket[2],4);
			// should fill in position at some point.
			if(Debug == DEBUG_DEVEL) {
				sprintf(DebugBuffer,"Got waypoint id: %.4x",wp.entityID);
				server->GamePrintString(DebugBuffer);
			}
			if(wp.entityID == 0) {
				server->GamePrintString("WPID was 0!");
			}
		}
	}

	return aLen;

}

// is run ~every 1/10th a second
DWORD EXPORT OnGameTimerTick(void) {
	
	tick++;
	if(state == STATE_NONE) {
		tick = 0;
	} else if(tick > 100) { // if we have been in a state for more then 10 seconds something is wrong
		server->GamePrintString("I am lost.  Disabling shopbot");
		state = STATE_NONE;
	} else {
		switch(state) {
			case STATE_ANTIIDLE:
				// just disable bot for now
				state = STATE_NONE;
				break;
			case STATE_WPFROMTOWN:	// leave town.  currently hardcoded to to wp to frigid highlands
				if(tick > 3 && wp.entityID != 0) {
					if(Debug) {
						server->GamePrintString("Taking WP out of town");
					}
					TakeWayPoint(wp.entityID,WAYPOINT_ACT5_2);
					// next state is taking wp back to town.
					state = STATE_WPTOTOWN;
					wp.entityID = 0;
					tick = 0; // reset tick
				}
				break;
			case STATE_WPTOTOWN:
				if(wp.entityID != 0) {
					if(Debug) {
						server->GamePrintString("Taking WP to town");
					}
					TakeWayPoint(wp.entityID,WAYPOINT_ACT5_1);
					// next state is to start walking to npc
					state = STATE_WALKNPC_START;
					tick = 0;
				}
				
				break;
			case STATE_WALKNPC_START:
				if(npc.npcID != 0) {
					if(Debug) {
						server->GamePrintString("Walking to NPC");
					}
					currentStep = path->GetFirstItem();
					// start walking
					state = STATE_WALKNPC;
					tick = 0;
				} else {
					// dont have an npc!?
					server->GamePrintString("Dont have an NPC!, shopbot disabled");
					state = STATE_NONE;
				}
				break;
			case STATE_WALKNPC:
				if(currentStep != NULL) {
					if(tick > 2) { // dont try walking too fast, should make this adjustable
						LOCATIONSTRUCT *loc;
						loc = (LOCATIONSTRUCT*)currentStep->lpData;
						if(Debug == DEBUG_DEVEL) {
					//		server->GamePrintString("Taking Step to NPC");
						}
						TakeStep(loc->x,loc->y);
						currentStep = path->GetNextItem(currentStep);
						
						// last point in our recorded path, try talking to where the NPC is located
						if(currentStep == NULL) {
							if(Debug == DEBUG_DEVEL) {
								server->GamePrintString("Taking Last Step to NPC");
							}
							TakeStep(npc.pos.x,npc.pos.y);
							state = STATE_OPENSTORE;
						}
						// reset tick
						tick = 0;
					}		
				}
				break;
			case STATE_OPENSTORE:
				if(tick > 3) {
					server->GamePrintString("Opening Store with NPC");
					OpenStore(npc.npcID,npc.pos.x,npc.pos.y);
					state = STATE_CLOSESTORE;
					tick = 0;
				}
				break;
			case STATE_CLOSESTORE:
				if(tick > 5) { // wait ~1 second before closing store to allow packets to get here
					if(Debug == DEBUG_DEVEL) {
						server->GamePrintString("Closing Store with NPC");
					}
					CloseStore(npc.npcID);
					state = STATE_NONE;
					tick = 0;
				}
			case STATE_WALKWP_START:  // basicly the same as walking to npc, but we traverse the linkedlist backwords
				if(wp.entityID != 0) {
					if(Debug) {
						server->GamePrintString("Walking to WP");
					}
					currentStep = path->GetLastItem();
					// start walking
					state = STATE_WALKWP;
					tick = 0;
				} else {
					// dont have an npc!?
					server->GamePrintString("Dont have an WP!, shopbot disabled");
					state = STATE_NONE;
				}
				break;
			case STATE_WALKWP:
				if(currentStep != NULL) {
					if(tick > 2) { // dont try walking too fast, should make this adjustable
						LOCATIONSTRUCT *loc;
						loc = (LOCATIONSTRUCT*)currentStep->lpData;
						if(Debug == DEBUG_DEVEL) {
						// server->GamePrintString("Taking Step to WP");
						}
						TakeStep(loc->x,loc->y);
						currentStep = path->GetPrevItem(currentStep);
						
						// first point in our recorded path 
						if(currentStep == NULL) {
							// TODO parse wp packet and make the last step location be the wp.
							if(Debug == DEBUG_DEVEL) {
								server->GamePrintString("Waiting on server->client Sync");
							}
							state = STATE_WPFROMTOWN;
						}
						// reset tick
						tick = 0;
					}		
				}
				break;
			default:
				break;
		}
	}
	return true;
}

BOOL LoadConfig(void) {
	char ConfigFile[1024];
	char Sections[4096];
	char error[1024];
	char *rval;
	int length,counter;
	CONFIGSTRUCT *cfg;

	sprintf(ConfigFile,"%s\\%s",ConfigPath,"shopbot.ini");

	if (_access(ConfigFile, 0))
	{ 
		char error[1024];
		sprintf(error, "Unable to open Config: %s", ConfigFile);
		server->GamePrintError(error);
		return false;
	}

	// get all the different sections
	length = GetPrivateProfileSectionNames(Sections,4096,ConfigFile);
	counter = 0;
NEXT:
	while(counter < length) {
		cfg = new CONFIGSTRUCT;
		cfg->Description[0] = '\0';
		cfg->MagicPrefix[0] = '\0';
		cfg->MagicSuffix[0] = '\0';
		cfg->item.ItemCode[0] = '\0';

		if(strlen(&Sections[counter]) > 128 || Sections[counter] == '\0') {
			
			if(Debug) {
				sprintf(DebugBuffer,"LoadConfig: Bad Section length (%d)",strlen(&Sections[counter]));
				server->GamePrintInfo(DebugBuffer);
			}
			counter = counter + strlen(&Sections[counter]) + 1;
			delete cfg;
			goto NEXT;
		}

			
		strcpy(cfg->Section,&Sections[counter]);
		counter = counter + strlen(&Sections[counter]) + 1;
		// now query the section and fill in data
		
		// ItemCode should be 3 or 4 chars long
		rval = server->GetHackProfileString("shopbot", cfg->Section, "Code");
		if(strlen(rval)) {
			if(strlen(rval) != 4 && strlen(rval) != 3) {
				sprintf(error,"Bad Code value '%s' in section [%s], Skipping",rval,cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			
			// if 4 char is a space, chop it off
			if(rval[3] == ' ') 
				rval[3] = '\0';

			strcpy(cfg->item.ItemCode,rval);
			// delete rval;
		} else {
			cfg->item.ItemCode[0] = '\0';
		}

		// Item Description
		rval = server->GetHackProfileString("shopbot", cfg->Section, "Description");
		if(strlen(rval)) {
			if(strlen(rval) > 127) {
				sprintf(error,"Description too long in section [%s], Skipping",cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			strcpy(cfg->Description,rval);
			// delete rval;
		} else {
			cfg->Description[0] = '\0';
		}

		rval = server->GetHackProfileString("shopbot", cfg->Section, "MagicPrefix");
		if(strlen(rval)) {
			if(strlen(rval) > 127) {
				sprintf(error,"MagicPrefix too long in section [%s], Skipping",cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			strcpy(cfg->MagicPrefix,rval);
			// delete rval;
		} else {
			cfg->MagicPrefix[0] = '\0';
		}

		rval = server->GetHackProfileString("shopbot", cfg->Section, "MagicSuffix");
		if(strlen(rval)) {
			if(strlen(rval) > 127) {
				sprintf(error,"MagicSuffix too long in section [%s], Skipping",cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			strcpy(cfg->MagicSuffix,rval);
			// delete rval;
		} else {
			cfg->MagicSuffix[0] = '\0';
		}
		// add to config linked list.
		config->AddItem(cfg);
		if(Debug == DEBUG_DEVEL) {
			sprintf(DebugBuffer,"[%s] des='%s' pre='%s' code='%s' suf='%s'",cfg->Section,cfg->Description,cfg->MagicPrefix,cfg->item.ItemCode,cfg->MagicSuffix);
			server->GamePrintString(DebugBuffer);
		}
	}
	return true;
}

CONFIGSTRUCT * ItemMatchCheck(ITEMSTRUCT *item) {
	LinkedItem_t *cfgItem;
	CONFIGSTRUCT *cfg;

	BOOL match;
	for(cfgItem = config->GetFirstItem();cfgItem != NULL;cfgItem = config->GetNextItem(cfgItem)){
		match = true;
		cfg = (CONFIGSTRUCT*)cfgItem->lpData;

		if(cfg->item.ItemCode[0] != '\0' && strcmpi(cfg->item.ItemCode,item->ItemCode) != 0) 
			match = false;

		if(cfg->MagicPrefix[0] != '\0') {
			if(item->MagicPrefix != NULL) {
				if(strcmpi(cfg->MagicPrefix,item->MagicPrefix->Prefix) != 0) {
					match = false;
				}
			} else {
				match = false;
			}
		}
		if(cfg->MagicSuffix[0] != '\0') {
			if(item->MagicSuffix != NULL) {
				if(strcmpi(cfg->MagicSuffix,item->MagicSuffix->Suffix) != 0) {
					match = false;
				}
			} else {
				match = false;
			}
		}
		if(match) {
			return cfg;
		}

	}
	return NULL;
}

// 
// stuff below here is from ClientCore.cpp
//////////////////////////////////////////////////////////////////////
// Dll entry/exit
//////////////////////////////////////////////////////////////////////
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
	BOOL hResult = TRUE;
	char *t;
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			
			// Create server struct			
			server = new FUNCTIONENTRYPOINTS; 

			// Bind exported functions from server
			HMODULE hModule;
			hModule = (HMODULE)GetModuleHandle("D2HackIt");
			
			// Macros make this look a lot nicer :)
			BIND_TO_SERVER(GamePrintInfo);
			BIND_TO_SERVER(GamePrintError);
			BIND_TO_SERVER(GamePrintVerbose);
			BIND_TO_SERVER(GamePrintString);
			BIND_TO_SERVER(GameCommandLine);
			BIND_TO_SERVER(GameSendPacketToServer);
			BIND_TO_SERVER(GameSendPacketToGame);
			BIND_TO_SERVER(GetFingerprint);
			BIND_TO_SERVER(Intercept);
			BIND_TO_SERVER(GetHackProfileString);
			BIND_TO_SERVER(SetHackProfileString);
			BIND_TO_SERVER(GetThisgameStruct);

			// Get plugin path
			t = ConfigPath;
			if (GetModuleFileName((HINSTANCE)hModule, t, _MAX_PATH)) {
				int p=strlen(ConfigPath);
				while (p) {
						if (ConfigPath[p] == '\\')
							{ ConfigPath[p] = 0; p=0;}
						else
					p--;
				}
			}	 
			// initiate client
			// hResult = OnClientStart();
			break;

		case DLL_PROCESS_DETACH:
			// kill client

			// hResult = OnClientStop();

			delete server;
			break;
    } 
    return hResult;
}

//////////////////////////////////////////////////////////////////////
// Stubfunctions for 'property get' functions.
//////////////////////////////////////////////////////////////////////
LPCSTR	EXPORT GetModuleAuthor()		{return ModuleAuthor;}
LPCSTR	EXPORT GetModuleWebsite()		{return ModuleWebsite;}
DWORD	EXPORT GetModuleVersion()		{return ModuleVersion;}
LPCSTR	EXPORT GetModuleEmail()			{return ModuleEmail;}
LPCSTR	EXPORT GetModuleDescription()	{return ModuleDescription;}

//////////////////////////////////////////////////////////////////////
// OnClientCommandLine
// -------------------------------------------------------------------
// The modules own extension of the command line interface. Any custom
// commands you add are parsed here.
//
// Return value should be TRUE, but it is not used at this 
// time.
//
// Arguments when we get here:
// argv[0]			Name of module
// argv[1]			Name of command (If supplied)
// argv[2 ... n]	The rest
//
// Syntax in the game: .<module> <arguments>
//////////////////////////////////////////////////////////////////////
BOOL EXPORT OnGameCommandLine(char* argv[], int argc)
{
	// Check if user supplied anything at all, if not assume help...
	if (argc==1)
		argv[argc++]="help";


	MODULECOMMANDSTRUCT* mcs=ModuleCommands;

	while (mcs->szName) {
		if (!stricmp(mcs->szName, argv[1]))
			break;
		mcs++;
	}

	char *p,*t,*msg,*fMsg;
	fMsg=new char[256];
	//
	// Is this a built-in function ?
	if (mcs->szName) {
		//
		// If functions returns false, show usage help
		if (!mcs->pFunc(argv, argc)) {
			t=new char[strlen(mcs->szUsage)+1];
			server->GamePrintInfo("Usage:");
			sprintf((char*)t, "%s", mcs->szUsage);
			if (strlen((char*)t))
			{
				msg=p=t;
				while (*p != 0) {
					if (*p == '\n') 
					{
						*(p++) = 0;
						sprintf(fMsg, "c4.%s %s", argv[0], msg);
						server->GamePrintInfo((char*)fMsg);
					if (*p != 0)
						msg = p;
					} else
						p++;
					}
				sprintf(fMsg, "c4.%s %s", argv[0], msg);
				server->GamePrintInfo((char*)fMsg);
			}
			delete t;
		}
	} else {
	// Unknown command, show catch-all help phraze.
	t=new char[128];
	sprintf(t, "Unknown command c4'%s %s'c0 - try c4'.%s help'c0 to get help.",
		argv[0], argv[1], argv[0]);
	server->GamePrintError(t);
	delete t;
	}
	delete fMsg;
	return TRUE;
}


//////////////////////////////////////////////////////////////////////
// OnCommandHelp
// -------------------------------------------------------------------
// Our default help function.
//
// Syntax in the game: .<module> <arguments>
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnCommandHelp(char** argv, int argc)
{
	// If command line is longer than 2, show usage for 'help'
	if (argc>2) return FALSE;

	char t[1024];
	sprintf(t, "Available commands for %s:", argv[0]);
	server->GamePrintInfo(t);

	// Loop through ModuleCommands[] and print their names
	for (int i=0; ModuleCommands[i].szName != NULL; i++)
	{
		sprintf(t, "c4.%s %s", argv[0], ModuleCommands[i].szName);
		server->GamePrintInfo(t);
	}

	sprintf(t, "For help on a specific command, type c4.%s <command> help", argv[0]);
	server->GamePrintInfo(t);
	return TRUE;
}
